;
;
;   Keyboard_Switch
;
; a PS/2 Keyboard switch for up to 4 PCs sharing 1 Keyboard
;
; written by Klaus Dormann, K@2m5.de
;
; Version: 12.11.2008 - stable release
;
; ATTINY2313 fuse settings (if other then default):
;   8Mhz internal RC Osc (7,3728 MHz crystal for debug because RS232 is used)
;   Startup: 14 CK + 65ms
;   BOD enabled - 2,7V
;   CKDIV8 disabled !!!!
;
; related docs:
;   keyboard switch schematics
;
; description of hardware:
;   portb0-7  Keyboard input of PCs
;             !  7  !  6  !  5  !  4  !  3  !  2  !  1  !  0  !
;             !data !clock!data !clock!data !clock!data !clock!
;             !    PC4    !    PC3    !    PC2    !    PC1    !
;
;   portd0,1  RS232 (debug) 38400 8n1
;   portd2,4  optional secondary input for mouse or 2nd keyboard (not implemented)
;   portd3,5  Keyboard input - clock on 3 with INT1 enabled, data on 5 
;   portd6    LED indicates current host selection = # of flashes
;   porta0,1  selected host output for future expansion - not with crystal (debug)
;             output can be used to switch mouse, VGA with a secondary cicuit   
;
; Changes:
;   05.11.2008 Changed from static (timer based 1ms/host * 4 hosts) port scheduling
;              to dynamic (ASAP with 3.5ms/frame). This reduces the risk of a host
;              function timeout and a race condition on early terminated frames.
;   07.11.2008 debug: indicate sent to #:off instead not sent
;   12.11.2008 added output of selected host to porta0,1 for further expansion 
 
            .NOLIST
            .INCLUDE "tn2313def.inc"
            .LIST
                   
.macro      addi
            subi  @0,-@1      ; subtract the negative of an immediate value
.endmacro
.macro      PrStr             ; print string from EEPROM to RS232
            ldi   zl,@0
            rcall pstr
.endmacro
;
; reserved registers
;
; (I): used during interrupt, (C): interrupt to program communication
; 
; 0x00 non immediate
.DEF  r2_addr     =r0         ; rs232 next write to address (debug)
.DEF  seq_bkup    =r1         ; sequencer resend step (active host only)
.DEF  hst_tmout   =r2         ; clock low timeout (host port stuck detection)
.DEF  hst_prty    =r3         ; (I) host xfer - parity generator counter
; 0x04
.DEF  hst_sel     =r4         ; (C) currently selected host (0-3)
.DEF  cur_host    =r5         ; currently active host (timeslice)
.DEF  hst_maskd   =r6         ; (C) current host data port mask
.DEF  hst_maskc   =r7         ; (C) current host clock port mask
; 0x08
.DEF  kbd_data    =r8         ; (C) keyboard xfer - data received/send
.DEF  kbd_prty    =r9         ; (I) keyboard xfer - parity check counter
.DEF  hq_data     =r10        ; host queue - data to be send to the host
.DEF  kbd_seq     =r11        ; keyboard init sequencer
; 0x0c
.DEF  c           =r12        ; general purpose register (temp2)
.DEF  k           =r13        ; (I) interrupt GPR (tempi2)
.DEF  s           =r14        ; (I) saved SREG
.DEF  r2_in       =r15        ; rs232 input convert to hex (debug)
; 0x10 immediate
.DEF  a           =r16        ; general purpose register (temp)
.DEF  i           =r17        ; (I) interrupt GPR (tempi)
.DEF  kbd_stat    =r18        ; (C) keyboard xfer state - +0x81 on error,
                              ; +0x41 on timeout - states:
                              ;   receive: 0 start, 1-8 data, 9 parity, 10 stop
                              ;   send: 0-7 data, 8 parity, 9 stop, 10 ack
                              ;   all: 11 complete
.DEF  kbd_send    =r19        ; data to be send to the keyboard
; 0x14
.DEF  hst_data    =r20        ; (C) host xfer - data send/received
.DEF  hst_seq     =r21        ; current step of the host port sequencer
                              ;   step 0-5  keyboard to host send
                              ;   step 6-22 keyboard emulation
.DEF  hst_stat    =r22        ; (C) host xfer state - +0x81 on error - states:
                              ;   even=falling clock odd=rising clock
                              ;     send/receive: 1,3,5,7,9,11,13,15 data,
                              ;     17 parity, 19 stop/ACK, 21 last clock,
                              ;     22 complete
.DEF  tim1_stat   =r23        ; (C) periodic LED indicator
                              ;   indicates the active host (hst_sel) by # of flashes
                              ;   >0x80 - 8Hz flashing for 1 second per each 
                              ;      0x10 added
; 0x18 immediate word
.DEF  hst_time    =r24        ; host frame timer - loaded with ms_cnth at start of frame
.DEF  unused_r25  =r25        ; unused
;     x                         unused
; 0x1c
;     y                       ; RAM address pointer
;     z                       ; Keyboard/Host sequencer pointer,
                              ; RS232 message pointer (debug),


;
; IO registers
;
; 0x30,0x31
.EQU  kbd_in      =pind       ; input from keyboard
.EQU  kbd_out     =ddrd       ; -output to keyboard (inverted - open collector driver)
.EQU  kclock      =3          ;   port d3 clock
.EQU  kdata       =5          ;   port d5 data

; 0x30-0x32
.EQU  led_tog     =pind       ; LED toggle
.EQU  led_dir     =ddrd       ; LED direction
.EQU  led_out     =portd      ; LED output
.EQU  stat_led    =6          ;   port d6 selected host indicator

; 0x33
.EQU  kbd_flgs    =gpior0     ; keyboard flags
.EQU  kbreak      =0          ;   bit0 break key (0xf0) / -make
.EQU  kextnd      =1          ;   bit1 extended keycode (0xe0) / -normal kc
.EQU  kcaps       =2          ;   bit2 caps mode
.EQU  ksend       =3          ;   (C) bit 3 send / -read
.EQU  ktype       =4          ;   bit4 next argument is typematic
.EQU  kled        =5          ;   bit5 next argument is LEDs
.EQU  kinit       =6          ;   bit6 host switch - reset pending

; 0x34
.EQU  hst_flgs     =gpior1    ; host flags
.EQU  hrecv       =0          ;   bit0 host com receive / -send
.EQU  hrwait      =1          ;   bit1 sequencer receive wait
.EQU  hstuck      =2          ;   bit2 host port stuck (data or clock low)
.EQU  hframe      =3          ;   bit3 host frame timer active (3,5ms)
.EQU  h4ms        =4          ;   bit4 host 4ms toggle (host port timeout)

; 0x35
.EQU  r2_flgs     =gpior2     ; debug flags
.EQU  rwrite      =0          ;   bit0 next hex input is write/-address
.EQU  rst_flgs    =gpior2     ; reset flags during init - bits equal to mcusr

; 0x36,0x37
.EQU  host_in     =pinb       ; input from hosts
.EQU  host_out    =ddrb       ; -output to hosts (inverted - open collector driver)

; 0x3a,0x3b
.EQU  hsel_init   =ddra       ; init ports 0,1 for host select propagation
.EQU  hsel_out    =porta      ; output hst_sel to porta0,1

; 0x4c,0x4d
.EQU  ms_cntl     =tcnt1l     ; ms counter - 1s per count, 65,536ms per rev
.EQU  ms_cnth     =tcnt1h
; 0x4a,0x4b
.EQU  ms_tmoutl   =ocr1al     ; 2ms keyboard clock timeout
.EQU  ms_tmouth   =ocr1ah
; 0x58,0x59
.EQU  ms_flgs     =tifr       ; ms counter flags
.EQU  ms_ena      =timsk      ; ms interrupt enables
.EQU  msena       =0b01000000 ;   timeout interrupt enable mask (OCIE1A)
.EQU  msdis       =0b10111111 ;       "       "    disable  "      "

; 0x52
.EQU  us_cnt      =tcnt0      ; s counter - 1s per count, 256s per rev
; 0x56
.EQU  us_host     =ocr0a      ; next host clock interrupt
; 0x5c
.EQU  us_dly      =ocr0b      ; next delay flag
; 0x58,0x59
.EQU  us_flgs     =tifr       ; s counter flags
.EQU  udelay      =0b00000100 ;   delay expired mask (OCF0B)
.EQU  us_ena      =timsk      ; s interrupt enables
.EQU  usena       =0b00000001 ;   host clock interrupt enable mask (OCIE0A)
.EQU  usdis       =0b11111110 ;    "     "       "    disable  "      "
               
;
; SRAM usage
;
.DSEG

hst_save:   .BYTE 24          ; host port regs save area (inactive host)
; 0x60 hq_data    (+0)
; 0x64 hst_stat   (+4)
; 0x68 hst_seq    (+8)
; 0x6c hst_flgs   (+12)
; 0x70 hst_tmout  (+16)
; 0x74 hst_time   (+20)

type_sav:   .BYTE 4           ; typematic settings for each host
led_sav:    .BYTE 4           ; LEDs lit on keyboard

.CSEG

;
; constants
;

;.EQU  debug       =1          ; uncomment for debug version (RS232 with crystal)
.IFNDEF debug
;.EQU  hst_sel_out =1          ; uncomment for host selection output on porta0:1
                              ; not available for units with external crystal (debug)!
.ENDIF

;scancodes to switch host port (only base codes allowed)
;
.EQU  hotkey      =0x58       ;[CAPS]
.EQU  sendcaps    =0x0e       ;[ ~ ] (US/UK) or [ ^ ] (DE)
.EQU  host1       =0x16       ;[ 1 ]
.EQU  host2       =0x1e       ;[ 2 ]
.EQU  host3       =0x26       ;[ 3 ]
.EQU  host4       =0x25       ;[ 4 ]
;alternative scancodes
;.EQU  hotkey      =0x58       ;[CAPS]
;.EQU  sendcaps    =0x76       ;[ESC]
;.EQU  host1       =0x05       ;[ F1 ]
;.EQU  host2       =0x06       ;[ F2 ]
;.EQU  host3       =0x04       ;[ F3 ]
;.EQU  host4       =0x0c       ;[ F4 ]

;
; reset and interrupt vectors
;
            rjmp  reset       ; Reset handler
            rjmp  illegal     ; INT0 handler
            rjmp  kbd_clock   ; INT1 handler
            rjmp  illegal     ; Timer1 capture event
            rjmp  kbd_timout  ; Timer1 compare match A 
            rjmp  tim_65ms    ; Timer1 overflow handler
            rjmp  illegal     ; Timer0 overflow handler
            rjmp  illegal     ; USART - RX complete
            rjmp  illegal     ; USART - data register empty
            rjmp  illegal     ; USART - TX complete
            rjmp  illegal     ; Analog comparator
            rjmp  illegal     ; Pin Change Interrupt
            rjmp  illegal     ; Timer1 compare match B
            rjmp  host_clock  ; Timer0 compare match A
            rjmp  illegal     ; Timer0 compare match B
            rjmp  illegal     ; USI start
            rjmp  illegal     ; USI overflow
            rjmp  illegal     ; EEPROM ready
            rjmp  illegal     ; Watchdog tripped

;
;
; ******************* Interrupt Routines ****************
;
;
; an illegal interrupt has occured
;
illegal:    reti                    ;for future use - ignore for now

;
; ******************* INT1 ****************
;
; Clock from keyboard
; sample/send data on low clock (called at falling edge of clock)
;
kbd_clock:  sbic  kbd_out,kclock
            reti                    ;busy forced by main
            in    s,sreg
            tst   kbd_stat          ;clock start?
            brne  kbdc_mode

            in    k,ms_cntl         ;2ms to timeout
            in    i,ms_cnth
            addi  i,8               ;8 * 256s
            out   ms_tmouth,i
            out   ms_tmoutl,k
            ldi   i,msena           ;clr and enable 2ms interrupt
            out   ms_flgs,i
            in    i,ms_ena
            ori   i,msena
            out   ms_ena,i

kbdc_mode:  sbis  kbd_flgs,ksend    ;send?
            rjmp  kbdc_recv
            
; send to keyboard

            cpi   kbd_stat,9
            brsh  kbds_stp
            cpi   kbd_stat,8
            breq  kbds_par

            ror   kbd_data          ;send data
            brcs  kbds_one
            sbi   kbd_out,kdata     ;send 0
            rjmp  kbdc_ext
kbds_one:   cbi   kbd_out,kdata     ;send 1
            inc   kbd_prty
            rjmp  kbdc_ext  
            
kbds_par:   bst   kbd_prty,0        ;send odd parity            
            brtc  kbds_par1
            sbi   kbd_out,kdata
            rjmp  kbdc_ext
kbds_par1:  cbi   kbd_out,kdata
            rjmp  kbdc_ext

kbds_stp:   brne  kbds_ack          ;send stop (1)
            cbi   kbd_out,kdata
            rjmp  kbdc_ext

kbds_ack:   sbis  kbd_in,kdata      ;data low for ack
            rjmp  kbdc_eob
            rjmp  kbdc_err

; receive from keyboard

kbdc_recv:  cpi   kbd_stat,9        ;which bit?                 
            brsh  kbdr_par
            cpi   kbd_stat,0
            brne  kbdr_dta

            sbis  kbd_in,kdata      ;start bit should be 0
            rjmp  kbdc_ext
            rjmp  kbdc_err

kbdr_dta:   sbic  kbd_in,kdata      ;sample data
            sec
            ror   kbd_data
            sbic  kbd_in,kdata
            inc   kbd_prty          ;count bits for parity check
            rjmp  kbdc_ext                  
            
kbdr_par:   brne  kbdr_stp
            bst   kbd_prty,0        ;check parity
            sbis  kbd_in,kdata   
            brts  kbdc_ext
            sbic  kbd_in,kdata
            brtc  kbdc_ext
            rjmp  kbdc_err

kbdr_stp:   sbis  kbd_in,kdata      ;stop bit should be 1

; common exit

kbdc_err:   ori   kbd_stat,0x80     ;flag error
kbdc_eob:   sbi   kbd_out,kclock    ;hold clock (pull down)
            clr   kbd_prty
            in    i,ms_ena          ;disable timeout
            andi  i,msdis
            out   ms_ena,i

kbdc_ext:   inc   kbd_stat          ;and normal exit
            out   sreg,s
            reti

;
; timer 1 compare match A
; keyboard 2ms frame timeout
;
kbd_timout: in    s,sreg
            ori   kbd_stat,0x40     ;flag timeout
            rjmp  kbdc_eob          
            
;
; timer 0 compare match A
; send to / receive from host clock (30s clock phase)
;
host_clock: in    s,sreg
            sbic  hst_flgs,hrecv     ;send to host?
            rjmp  host_recv
;
; send to host
;
            sbrs  hst_stat,0        ;falling or rising clock
            rjmp  hsts_fall

            in    i,host_out        ;release clock high
            and   i,hst_maskd       ;leave data untouched
            out   host_out,i
            rjmp  hsts_phase
            
hsts_fall:  in    i,host_in         ;check clock held by host
            and   i,hst_maskc
            breq  hsts_err
            in    i,host_out        ;set clock low
            or    i,hst_maskc
            out   host_out,i

hsts_phase: in    i,us_cnt          ;next clock phase 30s
            addi  i,30
            out   us_host,i

            in    i,host_out        ;data forced low?
            and   i,hst_maskd
            brne  hsts_d_ok
            in    i,host_in         ;collision?
            and   i,hst_maskd
            breq  hsts_err

hsts_d_ok:  sbrs  hst_stat,0        ;continue on rise - data change
            rjmp  hsts_exit

hsts_data:  cpi   hst_stat,17
            brsh  hsts_prty

            lsr   hst_data          ;send data
            brcc  hsts_dat0

hsts_dat1:  out   host_out,zero     ;release data high
            inc   hst_prty
            rjmp  hsts_exit

hsts_dat0:  in    i,host_out        ;set data low
            or    i,hst_maskd
            out   host_out,i
            rjmp  hsts_exit

hsts_prty:  brne  hsts_stop
            
            sbrc  hst_prty,0        ;send odd parity
            rjmp  hsts_dat0
            rjmp  hsts_dat1

hsts_stop:  cpi   hst_stat,19
            breq  hsts_dat1         ;send stop (1)

hsts_end:   in    i,us_ena          ;complete - disable host interrupt
            andi  i,usdis
            out   us_ena,i
            clr   hst_prty

hsts_exit:  inc   hst_stat          ;exit - prepare next clock phase
            out   sreg,s
            reti

hsts_err:   ori   hst_stat,0x80     ;flag errors in bit 7 of the status
            out   host_out,zero     ;clear hung bits
            rjmp  hsts_end
;
; host receive
;
host_recv:  
            sbrs  hst_stat,0        ;falling or rising clock
            rjmp  hstr_fall

            in    i,host_out        ;release clock high
            and   i,hst_maskd       ;leave data untouched
            out   host_out,i
            rjmp  hstr_phase
            
hstr_fall:  in    i,host_in         ;check clock held by host
            and   i,hst_maskc
            breq  hstr_err
            in    i,host_out        ;set clock low
            or    i,hst_maskc
            out   host_out,i

hstr_phase: in    i,us_cnt          ;next clock phase 30s
            addi  i,30
            out   us_host,i

hstr_d_ok:  sbrs  hst_stat,0        ;continue on rise - data change
            rjmp  hstr_exit

hstr_data:  cpi   hst_stat,17
            brsh  hstr_prty

            clc
            in    i,host_in         ;get data input
            and   i,hst_maskd
            breq  hstr_dat0
            inc   hst_prty          ;its a one!
            sec
hstr_dat0:  ror   hst_data
            rjmp  hstr_exit

hstr_prty:  brne  hstr_stop

            in    i,host_in         ;get parity input
            and   i,hst_maskd
            breq  hstr_par0
            inc   hst_prty          ;its a one!
hstr_par0:  rjmp  hstr_exit         

hstr_stop:  cpi   hst_stat,19       
            brne  hstr_end

            in    i,host_in         ;stop (1)?
            and   i,hst_maskd
            brne  hstr_stop1

            ldi   i,0x80+19         ;flag bad stop bit
            mov   hst_prty,i

hstr_stop1: in    i,host_out        ;send ack
            or    i,hst_maskd
            out   host_out,i
            rjmp  hstr_exit

hstr_err:   ori   hst_stat,0x80     ;flag errors in bit 7 of the status
            
hstr_end:   out   host_out,zero     ;complete - release data high
            in    i,us_ena          ;disable host interrupt
            andi  i,usdis
            out   us_ena,i
            tst   hst_prty          ;test error flags
            brpl  hstr_end1
            mov   hst_stat,hst_prty
hstr_end1:  sbrs  hst_prty,0        ;flag bad (even) parity?
            ldi   hst_stat,0x80+17  
            clr   hst_prty

hstr_exit:  inc   hst_stat          ;exit - prepare next clock phase
            out   sreg,s
            reti

;
; timer 1 overflow
; periodic LED update (65ms)
;
; indicates the currently selected (active) host by the number of flashes
; flashes with 8Hz, if the tim1_stat register is loaded with a value > 0x7F
; for every 0x10 added to 0x7F there will be 1 second of flashes
;
tim_65ms:   sei                     ;low priority interrupt runs enabled
            push  a
            in    a,sreg
            push  a
            tst   tim1_stat
            brpl  tim_period
            sbi   led_tog,stat_led  ;toggle too show error condition
            dec   tim1_stat
            brpl  tim_ex3
            rjmp  tim_ex1
tim_period: cbi   led_out,stat_led
            ldi   a,1
            sbic  kbd_flgs,kcaps
            rjmp  tim_ex2
            mov   a,tim1_stat
            lsr   a
            brcc  tim_exit
            lsr   a
            brcs  tim_exit
            cp    hst_sel,a
            brlo  tim_exit
            sbi   led_out,stat_led
tim_exit:   dec   tim1_stat
            brne  tim_ex1
tim_ex3:    mov   a,hst_sel
            lsl   a
            lsl   a
            addi  a,12
tim_ex2:    mov   tim1_stat,a
            sbi   led_out,stat_led
tim_ex1:    pop   a
            out   sreg,a
            pop   a
            reti


;
; ******************* RESET ****************
;

reset:      ldi   a,ramend          ;set Stack-Pointer
            out   spl,a

            clr   yh                ;RAM pointer (adressable < 256)
            .DEF  zero  =r29        ;yh will always be zero

            in    a,mcusr           ;save reset status & clear
            out   rst_flgs,a
            out   mcusr,zero

            clr   kbd_stat          ;init for keyboard com
            clr   kbd_prty
            
            sbic  rst_flgs,extrf    ;preserve state if extrf
            rjmp  no_mclear
            
            ldi   yl,hst_save       ;clean memory if borf or porf
mclear:     st    y+,kbd_prty       ;zero - cannot be used here
            cpi   yl,ramend+1
            brlo  mclear          

no_mclear:  ldi   a,0b00001000      ;falling edge int1
            out   mcucr,a
            
            ldi   a,0b00000010      ;timer 0+1 as s counter, prescaler clock/8   
            out   tccr0b,a          
            out   tccr1b,a          
            ldi   a,0b10000000      ;enable TOIE1 (periodic LED update)
            out   timsk,a
            out   gimsk,a           ;INT1 enable           
            sbi   led_dir,stat_led  ;enable LED

            .IFDEF hst_sel_out
              ldi   a,3               ;enable output for selected host
              out   hsel_init,a
            .ENDIF
            
            clr   tim1_stat         
            inc   tim1_stat

            .IFDEF debug              ;RS232 config 8n1 38400
              ldi   a,0x0b            ;7.3728 MHz crystal
              out   ubrrl,a
              ldi   a,0b00011000      ;receive and transmit enable
              out   ucsrb,a
              clr   r2_in
              clr   r2_addr
              rcall init_msg
            .ENDIF

            ldi   hst_seq,0x80      ;init for host com
            ldi   hst_stat,22
            clr   hst_prty

            ldi   a,4
            mov   cur_host,a        ;port countdown loop
            ldi   yl,hst_save+3
            ldi   i,0b11000000      ;prepare port mask for power check

clr_sav:    dec   cur_host
            std   y+4,hst_stat
            std   y+8,hst_seq
            clr   a                 ;next flag - power on (not stuck)
            in    c,host_in         ;check data or clock high (powered)
            and   c,i
            brne  clr_sv1

            .IFDEF debug
              PrStr power_msg
              rcall stuck_rst         ;this port is off   
            .ENDIF

            ldi   a,0b00000100      ;set hstuck (power off)
            rjmp  clr_sv2
                  
clr_sv1:    sbis  rst_flgs,extrf    ;retain host selection
            mov   hst_sel,cur_host  ;selected host = lowest available  
clr_sv2:    out   hst_flgs,a
            std   y+12,a
            lsr   i
            lsr   i
            dec   yl
            cpi   yl,hst_save
            brsh  clr_sav

            .IFDEF hst_sel_out
              out   hsel_out,hst_sel
            .ENDIF
            
            sei
            
            sbis  rst_flgs,extrf
            rjmp  pre_main

            ldi   kbd_send,0xff     ;reset keyboard on reset switch
            sbi   kbd_flgs,kinit
            rcall kb_send
            ldi   a,5
            mov   kbd_seq,a
            
pre_main:   out   rst_flgs,zero
            rjmp  m_reset                       
;
; ****************** Main Program ************
;

main:       
            .IFDEF debug
              sbic  ucsra,rxc         ;check for rs232 input
              rcall rs232_in
            .ENDIF

            cpi   hst_stat,22
            brsh  m_switch

            rcall m_kbd             ;host com did not complete - only kbd
            rjmp  main

m_switch:   
            mov   yl,cur_host
            addi  yl,hst_save       ;save registers for previous host
            st    y,hq_data
            std   y+4,hst_stat
            std   y+8,hst_seq
            in    a,hst_flgs
            std   y+12,a
            std   y+16,hst_tmout
            std   y+20,hst_time
           
m_reset:    inc   yl
            andi  yl,3
            mov   cur_host,yl
            addi  yl,hst_save

            ld    hq_data,y         ;restore registers for current host
            ldd   hst_stat,y+4
            ldd   hst_seq,y+8
            ldd   a,y+12
            out   hst_flgs,a
            ldd   hst_tmout,y+16
            ldd   hst_time,y+20

            rcall m_kbd             ;handle keyboard

m_mask:     mov   a,cur_host
            sec                     ;set host masks
            clr   hst_maskc
            rol   hst_maskc
            tst   a
            breq  m_mask1
m_mask2:    lsl   hst_maskc
            lsl   hst_maskc   
            dec   a
            brne  m_mask2
m_mask1:    mov   hst_maskd,hst_maskc
            lsl   hst_maskd

            sbis  hst_flgs,hstuck
            rjmp  m_notstuck        ;recovering stuck ports

m_stuck:    in    a,host_in         ;clock stuck?
            and   a,hst_maskc
            breq  m_stk1
            in    a,host_in         ;data stuck?
            and   a,hst_maskd
            brne  m_stk2

m_stk1:     sbrc  hst_seq,7         ;host sequencer active?
            rjmp  main

            .IFDEF debug
              rcall stuck_rst       
            .ENDIF

            ldi   tim1_stat,0x9f    ;flash 2 seconds if stuck
            ldi   hst_seq,0x80      ;clear any queued port activity
            rjmp  main

m_stk2:     cbi   hst_flgs,hstuck   ;no longer stuck!

            .IFDEF debug
              PrStr power_msg
              rcall host_num
              rcall colon
              PrStr on_msg
            .ENDIF

m_notstuck: in    a,ms_cntl         ;update timeout every 4ms
            in    a,ms_cnth
            in    c,hst_flgs
            eor   a,c
            andi  a,16              ;h4ms (bit4) != ms_cnt 4ms
            breq  m_host

            eor   c,a               ;toggle 4ms
            out   hst_flgs,c

            inc   hst_tmout
            brne  m_host

            sbi   hst_flgs,hstuck   ;timout after full count (256*4ms)
            ldi   hst_seq,0x80      ;clear queue

            .IFDEF debug
              in    a,host_in 
              and   a,hst_maskd
              breq  m_pwr
              rcall stuck_err
              rjmp  main
m_pwr:        PrStr power_msg
              rcall stuck_rst
            .ENDIF

main_x:     rjmp  main

m_host:     sbic  hst_flgs,hrecv    ;handle host port
            rcall hs_func           ;data received, process function

            sbis  hst_flgs,hframe   ;frame timer active?
            rjmp  m_hst1

            in    a,ms_cntl         ;frame timer
            in    a,ms_cnth
            sub   a,hst_time
            cpi   a,14              ;~3,5ms? (14*256s)
            brlo  main_x            ;frame timer not expired

            cbi   hst_flgs,hframe

m_hst1:     in    a,ms_cntl         ;preload for next frame
            in    hst_time,ms_cnth

            in    a,host_in         ;test clock hold
            and   a,hst_maskc
            brne  pc+2
            rjmp  main
            
            clr   hst_tmout
            in    a,host_in         ;test host request
            and   a,hst_maskd
            breq  m_recv            ;receive from host

            sbis  hst_flgs,hrwait
            rcall hs_actv
            rjmp  main

m_recv:     rcall hs_recv

            rjmp  main
            
;
; handle keyboard
;
m_kbd:      cpi   kbd_stat,11       ;keyboard buffer full or error?
            brlo  m_kbdx
            sbis  kbd_flgs,ksend
            rcall kb_recv
            cpi   kbd_stat,11
            brlo  m_kbdx
            sbic  kbd_flgs,ksend
            rcall kb_sndx
m_kbdx :    ret

;
; send to keyboard
;

kb_send:    sbi   kbd_out,kclock    ;force keyboard on hold
            rcall dly100us
            cbi   kbd_flgs,kbreak   ;set send, clear the rest
            cbi   kbd_flgs,kextnd
kb_snd1:    sbi   kbd_flgs,ksend    ;from kbd recv NAK
            mov   kbd_data,kbd_send
            clr   kbd_stat

            mov   a,kbd_data        ;save typematic or led
            sbis  kbd_flgs,ktype
            rjmp  kb_sx1
            cbi   kbd_flgs,ktype
            cpi   a,128
            brsh  kb_sx2
            ldi   yl,type_sav
            rjmp  kb_sx3
kb_sx1:     sbis  kbd_flgs,kled     ;LEDs?
            rjmp  kb_sx2
            cbi   kbd_flgs,kled
            cpi   a,8
            brsh  kb_sx2
            ldi   yl,led_sav
kb_sx3:     add   yl,hst_sel
            st    y,a
kb_sx2:            

            .IFDEF debug
              mov   a,kbd_send        ;show data
              rcall PrintHex
              rcall fromto
              ldi   a,0x4b            ;K
              rcall PrintChr
              rcall space
              sbic  kbd_flgs,kinit
              PrStr emu_msg
            .ENDIF

            rcall dly_wait
            sbi   kbd_out,kdata     ;start bit
            cbi   kbd_out,kclock    ;wait for clocks
            ret

kb_sndx:
            .IFDEF debug
              cpi   kbd_stat,11
              breq  kbsx_exit         ;exit keyboard send
              PrStr fail_msg          ;on error
              mov   a,kbd_stat
              rcall PrintHex
             ;rcall qms_time
              rcall space
kbsx_exit:    sbis  kbd_flgs,kinit  
              rcall crlf
            .ENDIF

            cpi   kbd_stat,11       ;flag next receive if
            brne  kbsx_2 
               
            mov   a,kbd_send        ;LED or typematic
            cpi   a,0xed
            brne  kbsx_3
            sbi   kbd_flgs,kled     ;next is LED argument
kbsx_3:     cpi   a,0xf3
            brne  kbsx_2
            sbi   kbd_flgs,ktype    ;next is typematic argument
               
kbsx_2:     clr   kbd_stat
            cbi   kbd_flgs,ksend    ;clear send flag
            cbi   kbd_out,kclock
            ret
            
;
; Receive Keycode from Keyboard
;

kb_recv:    breq  kb_vld            ;valid code received

            rcall dly100us          ;ERROR - hold clock low for 100s

            .IFDEF debug
              rcall kcode_msg
              PrStr fail_msg          ;on error
              mov   a,kbd_stat
              rcall PrintHex
             ;rcall qms_time
              rcall space
            .ENDIF

            rcall dly_wait          ;100 us expired?
            cpi   kbd_stat,0x8b     ;need to send a NAK?
            brsh  kb_nak            ;on error on the last clock
            sbrs  kbd_stat,6        ;on timeout
            rjmp  kb_skip
kb_nak:     ldi   kbd_send,0xfe
            rjmp  kb_snd1

kb_vld:     sbic  kbd_flgs,kinit
            rjmp  kb_ini1
            mov   a,kbd_data        ;valid data received
            cpi   a,0xe0            ;extended?
            brne  kb_vld1
            sbi   kbd_flgs,kextnd   ;set extended code flag
            rjmp  kb_rlse
kb_vld1:    cpi   a,0xf0            ;break?
            brne  kb_vld2
            sbi   kbd_flgs,kbreak   ;set break flag
            rjmp  kb_rlse

kb_vld2:    sbic  kbd_flgs,kextnd
            rjmp  kb_vld3
            cpi   a,hotkey          ;caps?
            brne  kb_vld3
            sbis  kbd_flgs,kbreak   ;make caps
            sbi   kbd_flgs,kcaps
            sbis  kbd_flgs,kcaps
            rjmp  kb_2host
            sbic  kbd_flgs,kbreak   ;break caps
            cbi   kbd_flgs,kcaps
            rjmp  kb_skp2           ;caps not sent to host

kb_vld3:    cpi   a,0xaa            ;keyboard response?
            brlo  kb_vld4
            rjmp  kb_2host
kb_vld4:    sbis  kbd_flgs,kcaps
            rjmp  kb_2host
            sbic  kbd_flgs,kextnd
            rjmp  kb_skp2
            sbic  kbd_flgs,kbreak
            rjmp  kb_skp2
            clr   c
            cpi   a,host1           ;key (1) - select host 1
            breq  kb_hstsw
            inc   c
            cpi   a,host2           ;key (2) - select host 2
            breq  kb_hstsw
            inc   c
            cpi   a,host3           ;key (3) - select host 3
            breq  kb_hstsw
            inc   c
            cpi   a,host4           ;key (4) - select host 4
            brne  kb_vld5

kb_hstsw:                           ;host select

            .IFDEF debug      
              rcall kcode_msg
              PrStr hsel_msg
              mov   a,c
              addi  a,49
              rcall PrintChr
            .ENDIF 

            ldi   zl,hst_save+12    ;check port functioning
            add   zl,c
            ld    a,z
            sbrs  a,hstuck
            rjmp  kb_hsw1

            .IFDEF debug            ;port not available for selection
              rcall space
              PrStr fail_msg
              mov   a,c
              addi  a,0x31
              rcall PrintChr
              rcall colon
              PrStr off_msg
            .ENDIF

            ldi   tim1_stat,0x9f    ;2s error flash
            rjmp  kb_skip

kb_hsw1:    mov   hst_sel,c         ;new host selected - init pending

            .IFDEF hst_sel_out
              out   hsel_out,hst_sel
            .ENDIF
            
            sbi   kbd_flgs,kinit
            clr   kbd_seq
            rjmp  kb_seq
            
kb_ini1:    
            .IFDEF debug
              rcall kcode_msg
            .ENDIF

            inc   kbd_seq
            mov   a,kbd_data
            cpi   a,0xfa
            breq  kb_ini2
            clr   kbd_seq
kb_ini2:    rjmp  kb_seq
            
kb_skp2:
            .IFDEF debug
              rcall kcode_msg
            .ENDIF

            rjmp kb_skip

kb_vld5:    cpi   a,sendcaps        ;top left key (~) - send caps lock
            brne  kb_skp2
            cp    hst_sel,cur_host  ;currently active?
            brne  pc+2
            sbrs  hst_seq,7         ;send only if queue free
            ret

            .IFDEF debug
              rcall kcode_msg
              PrStr caps_msg
            .ENDIF

            ldi   a,hotkey
            mov   hq_data,a
            cbi   kbd_flgs,kcaps
            rjmp  kb_snd4

kb_2host:   cp    hst_sel,cur_host  ;currently active?
            brne  pc+2
            sbrs  hst_seq,7         ;send only if sequencer is free
            ret

            .IFDEF debug
              rcall kcode_msg
            .ENDIF

kb_snd2:    mov   hq_data,kbd_data  ;copy data and flags to sequencer
            mov   a,kbd_data
            clr   hst_seq
            cpi   a,0xaa            ;send response as data -break -extnd
            brsh  kb_snd3
kb_snd4:    in    hst_seq,kbd_flgs
            andi  hst_seq,3         ;break, extended
kb_snd3:    mov   seq_bkup,hst_seq  ;for a resend

            .IFDEF debug
              PrStr qued_msg
              rjmp  kb_nocr
            .ENDIF
            
kb_skip:    
            .IFDEF debug
             rcall crlf
kb_nocr:
            .ENDIF

            cbi   kbd_flgs,kbreak
            cbi   kbd_flgs,kextnd
kb_rlse:    clr   kbd_stat          ;release keyboard clock
            cbi   kbd_out,kclock
            ret
;
; Keyboard Initialization Sequencer
;
; runs the code segments to initialize the keyboard
; to the new host port settings
;
kb_seq:     
            .IFDEF debug
              rcall crlf
            .ENDIF

            ldi   zl,low(kq_tbl)    ;point to code index
            ldi   zh,high(kq_tbl)
            add   zl,kbd_seq
            adc   zh,zero
            ijmp

kq_tbl:     rjmp  kq_f4             ;enable
            rjmp  kq_f3             ;set typematic
            rjmp  kq_type           ;typematic value
            rjmp  kq_ed             ;set LEDs
            rjmp  kq_led            ;LEDs value
            rjmp  kq_end            ;turn of init mode

            rjmp  kb_rlse           ;reset - dummy op

kq_f4:      ldi   kbd_send,0xf4     ;enable keyboard
            rjmp  kb_send

kq_f3:      ldi   kbd_send,0xf3     ;set typematic
            rjmp  kb_send

kq_type:    ldi   yl,type_sav       ;set typematic value
            add   yl,hst_sel
            ld    kbd_send,y
            andi  kbd_send,127      ;valid typematic mask
            rjmp  kb_send


kq_ed:      ldi   kbd_send,0xed     ;set LEDs
            rjmp  kb_send

kq_led:     ldi   yl,led_sav        ;set LEDs value
            add   yl,hst_sel
            ld    kbd_send,y
            andi  kbd_send,7        ;valid LEDs mask
            rjmp  kb_send

kq_end:     cbi   kbd_flgs,kinit
            rjmp  kb_rlse 


;
; send to active host
;
hs_actv:    cpi   hst_stat,22       ;normal completion or resend?
            breq  hs_act2
            
            .IFDEF debug
              PrStr fail_msg          ;on error
              mov   a,hst_stat
              rcall PrintHex
             ;rcall qms_time
              rcall crlf
            .ENDIF
            
hs_act3:    ldi   hst_stat,22
            cp    hst_sel,cur_host
            brne  hs_act2

            .IFDEF debug
              PrStr rsnd_msg
            .ENDIF
            
            mov   hst_seq,seq_bkup  ;resend! backup -> sequencer

hs_act2:    sbrc  hst_seq,7
            ret                     ;nothing to send
;
; Host Port Sequencer
;
; runs the code segment to handle host I/O for the current timeslice
;
hs_seq:     ldi   zl,low(seq_tbl)   ;point to code index
            ldi   zh,high(seq_tbl)
            add   zl,hst_seq
            adc   zh,zero
            ijmp

seq_tbl:                      ;keyboard codes
            rjmp  sq_data     ;  0 send data          -> 5
            rjmp  sq_break    ;  1 send break         -> 0
            rjmp  sq_extnd    ;  2 send extended make -> 0
            rjmp  sq_exbrk    ;  3 send extended break-> 1
            rjmp  sq_caps     ;  4 not used
            rjmp  sq_stat     ;  5 ..wait for status  ->end

                              ;host funtions (emulation)
            rjmp  sq_ack      ;  6 all dummy functions
            rjmp  sq_stat     ;  7   wait for status  ->end
            rjmp  sq_ack      ;  8 get ID (F2)
            rjmp  sq_AB       ;  9
            rjmp  sq_83       ; 10                    -> 7
            rjmp  sq_wack     ; 11 set typematic (F3)
            rjmp  sq_type     ; 12                    -> 6
            rjmp  sq_wack     ; 13 set LEDs (ED)
            rjmp  sq_leds     ; 14                    -> 6
            rjmp  sq_ack      ; 15 reset (FF)
            rjmp  sq_AA       ; 16                    -> 7
            rjmp  sq_EE       ; 17 echo (EE)          -> 7
            rjmp  sq_wack     ; 18 set/query scancode
            rjmp  sq_scan     ; 19
            rjmp  sq_ack      ; 20
            rjmp  sq_02       ; 21                    -> 6
            rjmp  sq_nak      ; 22 nak response       -> 7


sq_data:    mov   hst_data,hq_data  ;send keycode               
            ldi   hst_seq,5         ;jump to stat
            rjmp  hs_send

sq_break:   ldi   hst_data,0xf0     ;send break
            dec   hst_seq           ;continue with data
            rjmp  hs_send

sq_extnd:   ldi   hst_data,0xe0     ;send extended
            ldi   hst_seq,0         ;jump to data
            rjmp  hs_send

sq_exbrk:   ldi   hst_data,0xe0     ;send extended
            ldi   hst_seq,1         ;jump to break
            rjmp  hs_send

sq_caps:    ; no longer used
            ;ldi   hst_data,hotkey   ;send make caps
            ;mov   hq_data,hst_data  ;save for break caps
            ;ldi   hst_seq,1         ;jump to break
            ;rjmp  hs_send

sq_stat:    ldi   hst_seq,0x80      ;flag end

            .IFDEF debug
              rcall crlf
            .ENDIF

            ret
                                    
sq_wack:    sbi   hst_flgs,hrwait   ;wait for argument
sq_ack:     ldi   hst_data,0xfa     ;send ack
            inc   hst_seq           ;next
            rjmp  hs_send

sq_nak:     ldi   hst_data,0xfe     ;send nak
            ldi   hst_seq,7         ;wait for stat
            rjmp  hs_send

sq_ab:      ldi   hst_data,0xab     ;ID 1
            inc   hst_seq           ;next send ID 2
            rjmp  hs_send

sq_83:      ldi   hst_data,0x83     ;ID 2
            ldi   hst_seq,7         ;wait for stat
            rjmp  hs_send

sq_aa:      ldi   hst_data,0xaa     ;reset
            ldi   hst_seq,7         ;wait for stat
            rjmp  hs_send

sq_ee:      ldi   hst_data,0xee     ;echo
            ldi   hst_seq,7         ;wait for stat
            rjmp  hs_send

sq_02:      ldi   hst_data,0xab     ;the only scanset supported
            ldi   hst_seq,7         ;wait for stat
            rjmp  hs_send

sq_type:    cpi   hst_data,128
            brsh  sq_nak2
            ldi   yl,type_sav       ;save typematic setting
            rjmp  sq_stor

sq_leds:    cpi   hst_data,8
            brsh  sq_nak2
            ldi   yl,led_sav        ;save led setting
sq_stor:    add   yl,cur_host
            st    y,hst_data
sq_ack2:    ldi   hst_seq,6         ;next send ack
            ret

sq_scan:    cpi   hst_data,0        ;scancode select/show
            breq  sq_shosca         ;only scancode 2
            cpi   hst_data,2
            breq  sq_ack2
sq_nak2:    ldi   hst_seq,22        ;next send nak
            ret

sq_shosca:  inc   hst_seq           ;next send 02
            ldi   hst_data,0xfa
            ret

;
; send to host
;
hs_send:    clr   hst_stat
            sbi   hst_flgs,hframe   ;3.5ms timer for next frame
            cli
            in    a,us_cnt          ;50ms to 1st clock
            addi  a,50
            out   us_host,a
            ldi   a,usena           ;start host interrupts
            out   us_flgs,a
            in    a,us_ena
            ori   a,usena
            out   us_ena,a
            out   host_out,hst_maskd ;set start bit low
            sei

            .IFDEF debug
              rjmp  hcode_msg
            .ENDIF

            ret
;
; host request - receive from host
;
hs_recv:    clr   hst_stat
            sbi   hst_flgs,hrecv
            sbi   hst_flgs,hframe   ;3.5ms timer for next frame
            cli
            in    a,us_cnt          ;50ms to 1st clock
            addi  a,50
            out   us_host,a
            ldi   a,usena           ;start host interrupts
            out   us_flgs,a
            in    a,us_ena
            ori   a,usena
            out   us_ena,a
            sei
            ret
;
; function received from host
;
hs_func:        
            .IFDEF debug
              cpi   hst_seq,5         ;fix for a henn and egg problem
              breq  hf_fix1           ;waiting for stat isues crlf too late
              cpi   hst_seq,7
              brne  hf_fix2
hf_fix1:      ldi   hst_seq,0x80
              rcall crlf              ;-----end fix------------------
hf_fix2:      sbic  hst_flgs,hrwait
              rcall crlf
              rcall host_num
              rcall colon
              mov   a,hst_data        ;data
              rcall PrintHex
             ;rcall qms_time
              rcall space
            .ENDIF
            
            cbi   hst_flgs,hrecv    ;clear receive state
            cpi   hst_stat,22
            breq  hf_exit

            cpi   hst_data,0        ;check for stuck data pin
            brne  hf_err
            cpi   hst_stat,0x94     ;stop low
            brne  hf_err
            in    a,host_in         ;still stuck?
            and   a,hst_maskd
            brne  hf_err
            sbi   hst_flgs,hstuck
            ldi   hst_seq,0x80

            .IFDEF debug
              rcall stuck_err
            .ENDIF
            
            rjmp  hf_stop
hf_err:
            .IFDEF debug
              PrStr fail_msg
              mov   a,hst_stat
              rcall PrintHex
             ;rcall qms_time
              rcall space
              rcall space
            .ENDIF

hf_nak:     ldi   hst_seq,22
hf_stop:    ldi   hst_stat,22
            cbi   hst_flgs,hrwait
            ret

hf_exit:    sbic  kbd_flgs,kinit    ;during init all hosts on emulation
            rjmp  hf_emul           
            cp    hst_sel,cur_host  ;emulation or keyboard?
            brne  hf_emul
             
            cpi   hst_data,0xfe
            brne  hf_ext1
            rjmp  hs_act3           ;we want to resend

hf_ext1:    cpi   hst_data,0xf0     ;only scancode 2 allowed
            breq  hf_emul

            .IFDEF debug
              PrStr qued_msg
            .ENDIF

            mov   kbd_send,hst_data ;send to keyboard
            rjmp  kb_send

hf_emul:                            ;emulation
            .IFDEF debug
              PrStr emu_msg
            .ENDIF

hf_conv:    mov   a,hst_data        ;convert to sequence code
            subi  a,0xEC
            brlo  hf_nofunc

            ldi   zl,low(func_lst*2)
            ldi   zh,high(func_lst*2)
            add   zl,a
            adc   zh,zero
            lpm   a,z
            
            mov   hst_seq,a         ;now working on function
            cbi   hst_flgs,hrwait
            ret

hf_nofunc:  sbis  hst_flgs,hrwait   ;part of a previous function?
            rjmp  hf_nak
            cbi   hst_flgs,hrwait
            rjmp  hs_seq

;
; List of supported functions and their emulation sequence codes
;
func_lst:   .DB               22,13,17,22       ;EC-EF
            .DB   18,22, 8,11, 6, 6, 6,22       ;F0-F7
            .DB   22,22,22,22,22,22, 6,15       ;F8-FF


; 100s delay setup - part of keyboard error handling (clock hold) 

dly100us:   in    a,us_cnt
            addi  a,110             ;+10% tolerance
            out   ocr0b,a
            ldi   a,udelay          ;clear timer 0 compare b
            out   us_flgs,a
            ret

; wait for expiration

dly_wait:   in    a,us_flgs         ;timer expired (OCF0B set)?
            andi  a,udelay               
            breq  dly_wait          
            ret

            .IFDEF debug      

; rs232 input handled here
;
; enter 2 hex characters to enter a Reg/SRAM address
; enter "w" + 2 hex characters to write to address
; enter "r" to obtain a dump of the REG/SRAM space

rs232_in:   in    a,udr
            cpi   a,0x72            ;r?
            breq  r_dump
            cpi   a,0x77            ;w?
            brne  rsi_hex

            sbi   r2_flgs,rwrite    ;"w" = write to memory
            ret

r_dump:     PrStr reg_msg           ;"r" = read register/SRAM Dump
            clr   yl
r_dump2:    mov   a,yl
            rcall PrintHex
            rcall colon            
r_dump1:    ld    a,y+
            rcall PrintHex
            ldi   a,3
            and   a,yl
            brne  r_dump1
            rcall space
            ldi   a,31
            and   a,yl
            brne  r_dump1
            rcall crlf
            cpi   yl,0xe0
            brlo  r_dump2
            ret

rsi_hex:    subi  a,48              ;Hex 0-F for address / write
            cpi   a,10
            brlo  rsi_hx1 
            subi  a,39
            cpi   a,16
            brlo  rsi_hx1 
            ret
rsi_hx1:    sbrc  r2_in,0           ;if previous com
            rjmp  rsi_2nd
            swap  a                 ;1st digit - move to Bits 4:7
            com   a                 ;flag 1st digit already loaded
            mov   r2_in,a
            ret
rsi_2nd:    com   r2_in
            or    r2_in,a

            rcall crlf
            sbic  r2_flgs,rwrite
            rjmp  rs_write
            PrStr addr_msg
            mov   a,r2_in
            cpi   a,0xe0
            brlo  rs_addr
            mov   a,r2_addr
rs_addr:    mov   r2_addr,a
            rjmp  rs_com

rs_write:   mov   yl,r2_addr
            st    y,r2_in
            cbi   r2_flgs,rwrite
            PrStr write_msg
            mov   a,r2_in
            rcall PrintHex
            PrStr to_msg
            mov   a,r2_addr
rs_com:     rcall PrintHex
            clr   r2_in
            rjmp  crlf
                       
;
; debug messages to RS232
;
stuck_err:  PrStr fail_msg
stuck_rst:  rcall host_num
            rcall colon
            ldi   zl,off_msg
            rjmp  pstr

kcode_msg:  ldi   a,0x4B            ;K
            rcall PrintChr
            rcall colon
            ldi   zl,extnd_msg
            sbic  kbd_flgs,kextnd   ;extended
            rcall pstr
            ldi   zl,break_msg
            sbic  kbd_flgs,kbreak
            rcall pstr
kb_show:    mov   a,kbd_data        ;show data
            rcall PrintHex
            rjmp space

hcode_msg:  mov   a,hst_data
            rcall PrintHex
            rcall fromto
            rcall host_num

space:      ldi   a,0x20            ;print space
            rjmp PrintChr

host_num:   mov   a,cur_host
            addi  a,0x31            ;signal host
            rjmp  PrintChr

colon:      ldi   a,0x3a
            rjmp PrintChr

fromto:     ldi   a,0x3e      ;print >
            rjmp PrintChr

crlf:       ldi   a,10        ;print carriage return, linefeed
            rcall PrintChr
            ldi   a,13
            rjmp PrintChr
                             
;qms_time:   ldi   a,0x40            ;@
;            rcall PrintChr
;            in    a,ms_cntl         ;time in 1/4 ms increments
;            in    a,ms_cnth

; Hex Output to RS232
PrintHex:   push  a
            swap  a
            rcall prtx_dgt
            pop   a
prtx_dgt:   andi  a,0x0f
            addi  a,48              ;ASCII "0"
            cpi   a,(48+10)         ;ASCII "A"
            brlo  PrintChr
            addi  a,7               
; Character output to RS232
PrintChr:   sbis  ucsra,udre
            rjmp  PrintChr
            out   udr,a
            ret
; String output to RS232

pstr:       out   eear,zl
            sbi   eecr,eere
            in    a,eedr
            tst   a
            brne  pstr_wait
            ret
pstr_wait:  rcall PrintChr
            inc   zl
            rjmp  pstr

init_msg:   ldi   zh,high(init_txt<<1)
            ldi   zl,low(init_txt<<1)
init_nxt:   lpm   a,z+
            tst   a
            brne  init_wait
            ret
init_wait:  rcall PrintChr
            rjmp  init_nxt
            
init_txt:  .DB 10,13,10,13,"2m5 Keyboard Switch (12.11.08)",10,13,0


            .ESEG

reg_msg:    .DB 10,13,"Reg/SRAM Dump",10,13,0
break_msg:  .DB "F0.",0
extnd_msg:  .DB "E0.",0
qued_msg:   .DB "-> ",0
emu_msg:    .DB "<> ",0
off_msg:    .DB "Off",10,13,0
on_msg:     .DB "On",10,13,0
power_msg:  .DB "Power ",0
write_msg:  .DB "Write ",0
to_msg:     .DB " to ",0
fail_msg:   .DB "** ERROR Status: ",0
rsnd_msg:   .DB "Resend ",0
hsel_msg:   .DB "Host select: ",0
caps_msg:   .DB "(Caps) ",0
addr_msg:   .DB "Address: ",0

            .ENDIF            ;end of rs232  
